# oops-repository is Copyright 2011 Canonical Ltd.
#
# Canonical Ltd ("Canonical") distributes the oops-repository source code under
# the GNU Affero General Public License, version 3 ("AGPLv3"). See the file
# LICENSE in the source tree for more information.

"""basic operations on oopses in the db."""

import json
import time
import uuid
import datetime
from hashlib import sha1, md5
import pycassa
from pycassa.cassandra.ttypes import NotFoundException

DAY = 60*60*24
MONTH = DAY*30

_connection_pool = None
def connection_pool(config):
    '''Caching constructor of pycassa.ConnectionPool'''
    global _connection_pool
    if _connection_pool:
        return _connection_pool
    pool_size = config.get('pool_size', '')
    max_overflow = config.get('max_overflow', '')
    # I have no idea why max_retries is evaluating as 0 when not set, but here
    # we are, brute forcing this.
    creds = {'username': config['username'], 'password': config['password']}
    _connection_pool = pycassa.ConnectionPool(config['keyspace'],
                                              config['host'],
                                              pool_size=pool_size,
                                              max_overflow=max_overflow,
                                              max_retries=5,
                                              timeout=30,
                                              credentials=creds)
    return _connection_pool

def prune(config):
    """Remove OOPSES that are over 30 days old."""
    pool = connection_pool(config)
    dayoops_cf = pycassa.ColumnFamily(pool, 'DayOOPS')
    oops_cf = pycassa.ColumnFamily(pool, 'OOPS')
    # Find days to prune
    days = set()
    prune_to = time.strftime('%Y%m%d', time.gmtime(time.time() - MONTH))
    for key, _ in dayoops_cf.get_range():
        if key < prune_to:
            days.add(key)
    if not days:
        return
    # collect all the oopses (buffers all in memory; may want to make
    # incremental in future)
    batch_size = 10000
    oopses = []
    for day in days:
        columns_found = batch_size
        start_col = ''
        while columns_found==batch_size:
            columns = dayoops_cf.get(day, column_start=start_col)
            columns_found = len(columns)
            for column, oopsid in columns.items():
                start_col = column
                oopses.append(oopsid)
    # Remove the oopses
    batch = oops_cf.batch()
    map(batch.remove, oopses)
    batch.send()
    # Clean out the days aggregates
    # Clean out the days index
    batch = dayoops_cf.batch()
    map(batch.remove, days)
    batch.send()

def insert(config, oopsid, oops_json, user_token=None, fields=None,
           proposed_pkg=False):
    """Insert an OOPS into the system.

    :return: The day which the oops was filed under.
    """
    # make sure the oops report is a json dict, and break out each key to a
    # separate column. For now, rather than worrying about typed column values
    # we just coerce them all to strings.
    oops_dict = json.loads(oops_json)
    assert isinstance(oops_dict, dict)
    insert_dict = {}
    for key, value in oops_dict.items():
        insert_dict[key] = json.dumps(value)
    return _insert(config, oopsid, insert_dict, user_token, fields,
                   proposed_pkg)

def insert_dict(config, oopsid, oops_dict, user_token=None, fields=None,
                proposed_pkg=False):
    """Insert an OOPS into the system.

    :return: The day which the oops was filed under.
    """
    assert isinstance(oops_dict, dict)
    return _insert(config, oopsid, oops_dict, user_token, fields,
                   proposed_pkg)

def _insert(config, oopsid, insert_dict, user_token=None, fields=None,
            proposed_pkg=False):
    """Internal function. Do not call this directly.

    :param oopsid: The identifier for this OOPS.
    :param insert_dict: A dictionary containing the data to associate this OOPS
        with.
    :param user_token: An identifier for the user who experienced this OOPS.
    :param fields: A dictionary containing keys to increment counters for.
    :param proposed_pkg: True if the report's Tags contain package-from-proposed
    :return: The day which the oops was filed under.
    """
    pool = connection_pool(config)
    dayoops_cf = pycassa.ColumnFamily(pool, 'DayOOPS')
    oops_cf = pycassa.ColumnFamily(pool, 'OOPS')
    useroops_cf = pycassa.ColumnFamily(pool, 'UserOOPS')
    systemoopshashes_cf = pycassa.ColumnFamily(pool, 'SystemOOPSHashes')
    dayusers_cf = pycassa.ColumnFamily(pool, 'DayUsers')
    # Retry counter increments when a timeout is encountered. This has the risk
    # of double-counting, but the only alternative is to under count.
    counters_cf = pycassa.ColumnFamily(pool, 'Counters',
                                       retry_counter_mutations=True)
    counters_proposed_cf = pycassa.ColumnFamily(pool, 'CountersForProposed',
                                                retry_counter_mutations=True)
    day_key = time.strftime('%Y%m%d', time.gmtime())
    now_uuid = uuid.uuid1()

    try:
        oops_cf.insert(oopsid, insert_dict)
    except pycassa.MaximumRetryException:
        # Thrift doesn't like big data. If this persists to be a problem, we
        # should adjust thrift_framed_transport_size_in_mb upwards or use
        # chunked_insert from lp:daisy.
        for key, value in insert_dict.iteritems():
            oops_cf.insert(oopsid, {key: value})

    automated_testing = False
    if user_token.startswith('deadbeef'):
        automated_testing = True

    dayoops_cf.insert(day_key, {now_uuid:oopsid})
    # Systems running automated tests should not be included in the OOPS count.
    if not automated_testing:
        # Provide quick lookups of the total number of oopses for the day by
        # maintaining a counter.
        counters_cf.add('oopses', day_key)
        if fields:
            for field in fields:
                counters_cf.add('oopses:%s' % field, day_key)
        if proposed_pkg:
            for field in fields:
                counters_proposed_cf.add('oopses:%s' % field, day_key)

    if user_token:
        useroops_cf.insert(user_token, {oopsid : ''})
        # Build a unique identifier for crash reports to prevent the same
        # crash from being reported multiple times.
        date = insert_dict.get('Date', '')
        exec_path = insert_dict.get('ExecutablePath', '')
        proc_status = insert_dict.get('ProcStatus', '')
        if date and exec_path and proc_status:
            crash_id = '%s:%s:%s' % (date, exec_path, proc_status)
            if type(crash_id) == unicode:
                crash_id = crash_id.encode('utf-8')
            crash_id = md5(crash_id).hexdigest()
            systemoopshashes_cf.insert(user_token, {crash_id : ''})
        # TODO we can drop this once we're successfully using ErrorsByRelease.
        # We'll have to first ensure that all the calculated historical data is
        # in UniqueUsers90Days.
        dayusers_cf.insert(day_key, {user_token : ''})
        if fields:
            for field in fields:
                dayusers_cf.insert('%s:%s' % (field, day_key), {user_token : ''})

    return day_key

def bucket(config, oopsid, bucketid, fields=None, proposed_fields=False):
    """Adds an OOPS to a bucket, a collection of OOPSes that form a single
    issue. If the bucket does not exist, it will be created.

    :return: The day which the bucket was filed under.
    """
    pool = connection_pool(config)
    bucket_cf = pycassa.ColumnFamily(pool, 'Bucket')
    daybucket_cf = pycassa.ColumnFamily(pool, 'DayBuckets')
    daybucketcount_cf = pycassa.ColumnFamily(pool, 'DayBucketsCount',
                                             retry_counter_mutations=True)
    oops_cf = pycassa.ColumnFamily(pool, 'OOPS')

    # Get the timestamp.
    try:
        timestamp = oops_cf.get(oopsid, column_count=1, include_timestamp=True)
        timestamp = timestamp.values()[0][1]
        day_key = time.strftime('%Y%m%d', time.gmtime(timestamp / 1000000))
    except NotFoundException:
        # Eventual consistency. This OOPS probably occurred today.
        day_key = time.strftime('%Y%m%d', time.gmtime())

    bucket_cf.insert(bucketid, {uuid.UUID(oopsid) : ''})
    daybucket_cf.insert((day_key, bucketid), {oopsid : ''})

    if fields is not None:
        resolutions = (day_key[:4], day_key[:6], day_key)
        # All buckets for the given resolution.
        for field in fields:
            for resolution in resolutions:
                # We have no way of knowing whether an increment has been
                # performed if the write fails unexpectedly (CASSANDRA-2495).
                # We will apply eventual consistency to this problem and
                # tolerate slightly inaccurate counts for the span of a single
                # day, cleaning up once this period has passed. This will be
                # done by counting the number of columns in DayBuckets for the
                # day and bucket ID.
                daybucketcount_cf.add(':'.join((field, resolution)), bucketid)
        for resolution in resolutions:
            daybucketcount_cf.add(resolution, bucketid)
    return day_key

def update_bucket_versions(config, bucketid, version,
                           release=None, oopsid=None):

    pool = connection_pool(config)
    if release:
        bv_full = pycassa.ColumnFamily(pool, 'BucketVersionsFull')
        bv_count = pycassa.ColumnFamily(pool, 'BucketVersionsCount')
        bv_day = pycassa.ColumnFamily(pool, 'BucketVersionsDay')

        # Use the current day, rather than the day of the OOPS because this is
        # specifically used for cleaning up counters nightly. If a very old
        # OOPS gets processed here, we should still clean it up when we're
        # handling the data for today.
        day_key = time.strftime('%Y%m%d', time.gmtime())

        o = uuid.UUID(oopsid)
        release = release.encode('ascii', errors='ignore')
        if version:
            version = version.encode('ascii', errors='ignore')

        # When correcting the counts in bv_count, we'll iterate
        # BucketVersionsDay for day_key. For each of these columns, we'll look
        # up the correct value by calling bv_full.get_count(...).
        bv_full.insert((bucketid, release, version), {o: ''})
        bv_day.insert(day_key, {(bucketid, release, version): ''})

        try:
            bv_count.add(bucketid, (release, version))
        except pycassa.MaximumRetryException:
            # We'll clean these up nightly, so ignore timeouts.
            pass
    else:
        bucketversions_cf = pycassa.ColumnFamily(pool, 'BucketVersions',
                                                 retry_counter_mutations=True)
        bucketversions_cf.add(bucketid, version)

def query_bucket_versions(config, bucketid):
    pool = connection_pool(config)
    bucketversions_cf = pycassa.ColumnFamily(pool, 'BucketVersions')
    bucketversions = bucketversions_cf.xget(bucketid)
    if not bucketversions:
        return None
    return bucketversions

def update_errors_by_release(config, oops_id, system_token, release):
    release = release.encode('utf8')
    pool = connection_pool(config)
    firsterror = pycassa.ColumnFamily(pool, 'FirstError')
    errorsbyrelease = pycassa.ColumnFamily(pool, 'ErrorsByRelease')
    systemsfor = pycassa.ColumnFamily(pool, 'SystemsForErrorsByRelease')

    today = datetime.datetime.today()
    today = today.replace(hour=0, minute=0, second=0, microsecond=0)
    try:
        first_error_date = firsterror.get(release, columns=[system_token])
        first_error_date = first_error_date[system_token]
    except NotFoundException:
        firsterror.insert(release, {system_token: today})
        first_error_date = today

    # We use the OOPS ID rather than the system identifier here because we want
    # each crash from a system to take up a new column in this column family.
    # Each one of those columns should be associated with the date of the first
    # error for the system in this release.
    # 
    # Remember, we're ultimately tracking errors here, not systems, but we need
    # the system identifier to know the first occurrence of an error in the
    # release for that machine.
    #
    # For the given release for today, the crash should be weighted by the
    # first time an error occurred in the release for the system this came
    # from.  Multiplied by their weight and summed together, these form the
    # numerator of our average errors per calendar day calculation.

    errorsbyrelease.insert((release, today), {oops_id: first_error_date})
    systemsfor.insert((release, today), {system_token: ''})

def update_bucket_metadata(config, bucketid, source, version, comparator, release=''):
    # We only update the first and last seen version fields. We do not update
    # the current version field as talking to Launchpad is an expensive
    # operation, and we can do that out of band.
    pool = connection_pool(config)
    bucketmetadata_cf = pycassa.ColumnFamily(pool, 'BucketMetadata')
    metadata = pycassa.util.OrderedDict()
    try:
        # TODO: Drop the FirstSeen and LastSeen fields once BucketVersionsCount
        # is deployed, since we can just do a get(column_count=1) for the first
        # seen version and get(column_reversed=True, column_count=1) for the
        # last seen version.
        # N.B.: This presumes that we are using the DpkgComparator which we
        # won't be when we move to DSE.
        metadata = bucketmetadata_cf.get(bucketid)
        lastseen = metadata.get('LastSeen', None)
        if (not lastseen or comparator(lastseen, version) < 0):
            metadata['LastSeen'] = version
        lastseen_release = metadata.get('LastSeenRelease', None)
        if (not lastseen_release or \
                (lastseen_release.split()[-1] < release.split()[-1])):
            metadata['LastSeenRelease'] = release
        firstseen = metadata.get('FirstSeen', None)
        if (not firstseen or comparator(firstseen, version) > 0):
            metadata['FirstSeen'] = version
        firstseen_release = metadata.get('FirstSeenRelease', None)
        if (not firstseen_release or \
                (release.split()[-1] < firstseen_release.split()[-1])):
            metadata['FirstSeenRelease'] = release
    except NotFoundException:
        metadata['FirstSeen'] = version
        metadata['LastSeen'] = version

    if release:
        k = '~%s:FirstSeen' % release
        firstseen = metadata.get(k, None)
        if (not firstseen or comparator(firstseen, version) > 0):
            metadata[k] = version
        k = '~%s:LastSeen' % release
        lastseen = metadata.get(k, None)
        if (not lastseen or comparator(lastseen, version) < 0):
            metadata[k] = version

    if metadata:
        metadata['Source'] = source
        bucketmetadata_cf.insert(bucketid, metadata)

def update_bucket_systems(config, bucketid, system, version=None):
    '''Keep track of the unique systems in a bucket with a specific version of
       software.'''
    if not system or not version:
        return
    version = version.encode('ascii', errors='ignore')
    if not version:
        # If all we had were unicode characters.
        return
    pool = connection_pool(config)
    bucketsystems_cf = pycassa.ColumnFamily(pool, 'BucketVersionSystems2')
    bucketsystems_cf.insert((bucketid, version), {system: ''})

def update_source_version_buckets(config, source, version, bucketid):
    pool = connection_pool(config)
    srcversionbuckets_cf = pycassa.ColumnFamily(pool, 'SourceVersionBuckets')
    srcversionbuckets_cf.insert((source, version), {bucketid: ''})

def update_bucket_hashes(config, bucketid):
    '''Keep a mapping of SHA1 hashes to the buckets they represent.
       These hashes will be used for shorter bucket URLs.'''
    pool = connection_pool(config)
    hashes_cf = pycassa.ColumnFamily(pool, 'Hashes')
    bucket_sha1 = sha1(bucketid).hexdigest()
    k = 'bucket_%s' % bucket_sha1[0]
    hashes_cf.insert(k, {bucket_sha1: bucketid})

